unit ZLIBArchiveGlobals;

interface

uses SysUtils, Classes;

resourcestring
   ArchiveFileNotFound = 'Archive file not found';
   UnableToOpenFile = 'Unable to open archive %s';
   VerifyingArchive = 'Verifying archive integrity';
   ArchiveCorrupted = 'The archive appears to be corrupted';
   ExtractingRegistry = 'Extracting registry';
   RegistryReadError = 'Error reading registry';
   NoArchiveOpen = 'No archive is currently open';
   IndexBounds = 'Index out of bounds';
   FilePathError = 'Error extracting file to %s';
   ExtractingFile = 'Extracting %s';
   SkippingFile = 'Skipping %s';
   FileCRC = '%s failed CRC check';
   FileNotFound = '%s not found in archive';
   FileAlreadyExists = '%s already exists in the archive';
   UpdatingFile = 'Updating %s';
   MovingTempFile = 'Moving and renaming the temporary file';
   WritingRegistry = 'Writing archive registry';
   OverwriteFile = 'Replace %s (Yes/No/Cancel/All):';
   OverwriteFile2 = 'Replace %s';
   ArchiveDataNotFound = 'No valid archive data found';
   ZLBArchiveVersionError = 'The archive was created with a different version of ZLBArchive (v%d)';

type
   TCRC32Table = array[0..255] of longint;

   TFileStatus = (fsCompressed, fsStored) ;
   TFileCompressionLevel = (fcNone, fcFastest, fcDefault, fcMaximum) ;
   TInternalFlags = (ifOpening, ifExtracting, ifAdding, ifUpdating, ifWritingRegistry, ifClear, ifVerifyingArchiveCRC, ifMovingTempFile) ;

   TFileAddUpdateEvent = procedure(var Path, Name: shortstring) of object;
   TReplaceFileEvent = procedure(Filename: string; var Replace: boolean) of object;
   TFileExtractedEvent = procedure(filename: string) of object;
   TErrorLogEvent = procedure(Msg: string) of object;
   TProgressEvent = procedure(Msg: string; Percentage: longint) of object;

   EFileCorruption = Exception;
   EFileNotFound = Exception;
   EArchiveRegistry = Exception;
   EGeneralArchiveException = Exception;
   EExtractionError = Exception;
   ENoArchive = Exception;
   EArchiveIndexBounds = Exception;
   EFileNotFoundInArchive = Exception;

   TZLBFileRec = record
      name: shortstring;
      path: shortstring;
      osize: longint;
      csize: longint;
      start: longint;
      CRC: longint;
      Status: TFileStatus;
      CompressionLevel: TFileCompressionLevel;
   end;

   TZLBArchiveRec = record
      CRC: longint;
      FileCount: longint;
      Version: longint;
      RegistrySize: longint;
      ArchiveSize: longint;
      CompressedRegistry: Boolean;
      Reserved: array[0..11] of byte;
   end;

var
   oCRC32Table: TCRC32Table;

Function InitCRC32:TCRC32Table;
function GetCRC32(DataStream: TStream; IgnoreCount: longint; StartPos: Cardinal = 0) : longint;

implementation

Function InitCRC32:TCRC32Table;
var
   crc, poly: longint;
   i, j: longint;
begin
//   fillchar(InitCRC32, sizeof(result) , 0) ;

   poly := longint($EDB88320) ;
   for i := 0 to 255 do
   begin
      crc := i;
      for j := 8 downto 1 do
      begin
         if (crc and 1) = 1 then
            crc := (crc shr 1) xor poly
         else
            crc := crc shr 1;
      end;
      InitCRC32[i] := crc;
   end;
end;

function GetCRC32(DataStream: TStream; IgnoreCount: Integer; StartPos: Cardinal = 0): longint;
var
   crc, checked, buffersize, fsize, count: longint;
   BufferArray: array[0..10239] of byte;
   originalposition: longint;
begin
   originalposition := datastream.Position;
   dataStream.Seek(StartPos, soFromBeginning);
   crc := longint($FFFFFFFF) ;
   fsize := (datastream.size-StartPos) - IgnoreCount;
   while True do
   begin
      if fsize <= 0 then break;
      if fsize >= 10240 then
         buffersize := 10240
      else
         buffersize := fsize;
      Count := DataStream.Read(BufferArray, BufferSize) ;
      checked := 0;
      while checked < Count do
      begin
         crc := ((crc shr 8) and $FFFFFF) xor oCRC32Table[(crc xor bufferArray[checked]) and $FF];
         inc(checked) ;
      end;
      dec(fsize, buffersize) ;
   end;
   result := (crc xor longint($FFFFFFFF) ) ;
   datastream.seek(originalposition, soFromBeginning) ;
end;

initialization
   oCRC32Table := InitCRC32;
end.
